﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Globalization;

namespace Plecak
{
    public class RucksackAlgorithm
    {
        public RucksackAlgorithm() { }

        public void Run() 
        {

		    var chromosome = RunAlgorithm(populationSize, maxNumberOfIteration, selectionCount, 
                                            mutationProbability, crosoverProbability);

		    // wartośc plecaka zapakowanego przez chromosome
		    double totalWorth = 0;
		    double totalWeight = 0;

		    for (int i = 0; i < objectData.Count; i++) 
            {
			    if (chromosome.Table[i] == 1) 
                {
				    totalWorth += objectData[i].Worth;
				    totalWeight += objectData[i].Weight;
			    }
		    }

		    Console.WriteLine("Maksymalna waga : " + maxRucksackWeight);
		    Console.WriteLine("Waga plecaka : " + totalWeight);
            Console.WriteLine("Wartość plecaka : " + totalWorth);
	    }

        private List<Chromosome> InitializePopulation(int populationSize, int countOfGenes)
        {
            var result = new List<Chromosome>();
            for (int i = 0; i < populationSize; i++)
            {
                result.Add(new Chromosome(countOfGenes));
            }
            return result;
        }

         // ocena populacji
        private void EvaluatePopulation(List<Chromosome> population) 
        {
		    foreach (var chromosome in population) 
            {
			    // oblicz wartosc przystosowania każdego osobnika
			    chromosome.Fitness = CalculateFitness(chromosome);
		    }

		    // posortuj osobniki wg wartośći przysotowania
            population.Sort();

            // ustaw osobniki w populacji od największej funkcji przystosowania do najmniejszej
            population.Reverse();
	    }

        private double CalculateFitness(Chromosome chromosome)
        {
            // wartośc plecaka zapakowanego przez chromosome
            double knapsackWorth = 0;
            double knapsackWeight = 0;

            for (int i = 0; i < objectData.Count; i++)
            {
                if (chromosome.Table[i] == 1)
                {
                    knapsackWorth += objectData[i].Worth;
                    knapsackWeight += objectData[i].Weight;
                }
            }

            if (knapsackWeight > maxRucksackWeight)
            {
                return knapsackWorth - knapsackWorth
                        * (knapsackWeight / maxRucksackWeight);
            }
            else
            {
                return knapsackWorth;
            }
        }

        private Chromosome RunAlgorithm(int populationSize, int maxNumberOfIteration, int selectionCount, 
                                        double mutationProbability, double crosoverProbability) 
        {
            // ilość przedmiotów do spakowania
		    var countOfObject = objectData.Count; 

		    var iteration = 0;
		    var stop = false;

		    Chromosome theBest = null;

		    // inicjalizacja populacji początkowej
		    var population = InitializePopulation(populationSize, countOfObject);

		    EvaluatePopulation(population);
		    theBest = population[0];

		    int iterationWithoutChangeTheBest = 0;
		    while (!stop) 
            {

			    var newPopulation = SelectionPopulation(population, selectionCount);

			    // krzyzowanie osobnikow
			    for (int i = 0; i < newPopulation.Count - 1; i = i + 2) 
                {
				    newPopulation[i].Crossover(newPopulation[i + 1], crosoverProbability);
			    }

			    foreach (var chromosome in newPopulation) 
                {
				    chromosome.Mutate(mutationProbability);
			    }

			    EvaluatePopulation(newPopulation);
			    var localTheBest = newPopulation[0];
			    if (localTheBest.Fitness > theBest.Fitness) 
                {

				    Console.WriteLine("-------------------------------------------------");
				    Console.WriteLine("Iteracja - " + (iteration+1));
				    Console.WriteLine("Przed   - ");
                    Console.WriteLine(theBest.Print());
				    Console.WriteLine("Po    - ");
                    Console.WriteLine(localTheBest.Print());

				    theBest = new Chromosome(localTheBest);
				    iterationWithoutChangeTheBest = 0;
			    } 
                else 
                {
				    iterationWithoutChangeTheBest++;
			    }

			    // czy wykonano maksymalną ilośc iteracji
			    if (iteration >= maxNumberOfIteration) 
                {
				    stop = true;
			    } 
                else 
                {
				    iteration++;

				    // brak poprawy wyniku przez kolejne 10 iteracji
				    if (iterationWithoutChangeTheBest > maxIterationWithoutChangeTheBest) 
                    {
					    stop = true;
				    }
			    }

		    }

            Console.WriteLine("Liczba iteracji " + iteration);

		    return theBest;

	    }

        private List<Chromosome> SelectionPopulation(List<Chromosome> population, int selectionCount)
        {
            var random = new Random();

            var newPopulation = new List<Chromosome>();

            for (int i = 0; i < population.Count; i++)
            {

                var temp = new List<Chromosome>();
                for (int j = 0; j < selectionCount; j++)
                {
                    temp.Add(population[random.Next(population.Count)]);
                }
                temp.Sort();
                newPopulation.Add(temp[temp.Count - 1]);
            }

            return newPopulation;
        }

        public void ReadProperties()
        {
            try
            {
                var br = new StreamReader("properties.txt");

                populationSize = int.Parse(br.ReadLine());
                maxNumberOfIteration = int.Parse(br.ReadLine());
                maxIterationWithoutChangeTheBest = int.Parse(br.ReadLine());
                
                // ilość osobnikow podczas wyboru selekcji turniejowej
                selectionCount = int.Parse(br.ReadLine());

                mutationProbability = double.Parse(br.ReadLine(), CultureInfo.InvariantCulture);
                crosoverProbability = double.Parse(br.ReadLine(), CultureInfo.InvariantCulture);

                br.Dispose();
            }
            catch (Exception e)
            {
                Console.WriteLine("Wystąpił błąd wczytywana ustawień algorytmu z pliku properties.txt");
                Console.WriteLine(e.StackTrace);
                Environment.Exit(1);
            }
        }

        public void ReadInputData()
        {
            try
            {
                var br = new StreamReader("input.txt");

                maxRucksackWeight = double.Parse(br.ReadLine(), CultureInfo.InvariantCulture);

                objectData.Clear();

                string line = null;
                while ((line = br.ReadLine()) != null)
                {
                    var tab = line.Split(' ');
                    var o = new ObjectData(double.Parse(tab[0], CultureInfo.InvariantCulture),
                                            double.Parse(tab[1], CultureInfo.InvariantCulture));
                    objectData.Add(o);
                }

                br.Close();

            }
            catch (Exception e)
            {
                Console.WriteLine("Wystąpił błąd wczytywana danych wejściowych aplikacji input.txt");
                Console.WriteLine(e.StackTrace);
                Environment.Exit(1);
            }
        }

        // obiekty wczytane z wejścia, które należy zapakować do plecaka
        private readonly List<ObjectData> objectData = new List<ObjectData>();

        // maksymalna waga plecaka
        private double maxRucksackWeight;

        // wielkość populacji
        private int populationSize;

        // maksymalna liczba itracji algorytmu
        private int maxNumberOfIteration;

        // maksymalna liczba iteracji bez zmiany najlepszego osobnika
        private int maxIterationWithoutChangeTheBest;

        // liczba osobniów wybieranych w operacji selekcji w metodzie turniejowej
        private int selectionCount;

        // prawdopodobieństwo mutacji
        private double mutationProbability;

        // prawdopodobieństwo krzyżowania
        private double crosoverProbability;
    }
}
